//
// tu58 - Emulate a TU58 over a serial line
//
// Original (C) 1984 Dan Ts'o <Rockefeller Univ. Dept. of Neurobiology>
// Update (C) 2005 Don North <ak6dn_at_mindspring_dot_com>
//
// This is the TU58 emulation program written at Rockefeller Univ., Dept. of
// Neurobiology. We copyright (C) it and permit its use provided it is not
// sold to others. Originally written by Dan Ts'o circa 1984 or so.
//
// Extensively rewritten to work reliably under the CYGWIN unix subsystem.
// Has been tested up to 38.4Kbaud hosted on a 1GHz Pentium3 Win2K platform.
// Added native windows serial comm support compile option (for break detect).
// Added multithreaded supervisor for command parser and emulator restart.
// Added timing delay option to allow emulator to pass diagnostic ZTUUF0.BIN.
// Supports four (or more, compile option) logical TU58 devices per controller.
//
// 12 Jul 2005 - donorth - Initial rewrite
// 25 Feb 2006 - donorth - Added XXDP init support; create zeros entire tape
// 25 Nov 2007 - donorth - Fixed use of lseek to detect seeking past LEOF
//

//
// TU58 File access routines
//


#include "common.h"



// file data structure

struct {
    int		fd;		// file descriptor
    char	*name;		// file name
    char	rflag : 1;	// read allowed
    char	wflag : 1;	// write allowed
    char	cflag : 1;	// create allowed
    char	iflag : 1;	// init RT-11 structure
    char	xflag : 1;	// init XXDP structure
} file [NTU58];

int fpt; // number of active file descriptors



//
// init file structures for all units
//
void fileinit (void)
{
    int unit;

    for (unit = 0; unit < NTU58; unit++) {
	file[unit].fd = -1;
	file[unit].name = NULL;
	file[unit].rflag = 0;
	file[unit].wflag = 0;
	file[unit].cflag = 0;
	file[unit].iflag = 0;
	file[unit].xflag = 0;
    }
    fpt = 0;
    return;
}



//
// close file structures for all units
//
void fileclose (void)
{
    int unit;

    for (unit = 0; unit < NTU58; unit++) {
	if (file[unit].fd != -1) {
	    close(file[unit].fd);
	    file[unit].fd = -1;
	}
    }
    return;
}



//
// init RT-11 file directory structures (based on RT-11 v5.3)
//
static int rt11_init (int fd)
{
    int i;

    static short boot[] = { // offset 0000000
	0000240, 0000005, 0000404, 0000000, 0000000, 0041420, 0116020, 0000400,
	0004067, 0000044, 0000015, 0000000, 0005000, 0041077, 0047517, 0026524,
	0026525, 0067516, 0061040, 0067557, 0020164, 0067157, 0073040, 0066157,
	0066565, 0006545, 0005012, 0000200, 0105737, 0177564, 0100375, 0112037,
	0177566, 0100372, 0000777
    };

    static short bitmap[] = { // offset 0001000
	0000000, 0170000, 0007777
    };

    static short direct1[] = { // offset 0001700
	0177777, 0000000, 0000000, 0000000, 0000000, 0000000, 0000000, 0000000,
	0000000, 0000001, 0000006, 0107123, 0052122, 0030461, 0020101, 0020040,
	0020040, 0020040, 0020040, 0020040, 0020040, 0020040, 0020040, 0020040,
	0042504, 0051103, 0030524, 0040461, 0020040, 0020040
    };

    static short direct2[] = { // offset 0006000
	0000001, 0000000, 0000001, 0000000, 0000010, 0001000, 0000325, 0063471,
	0023364, 0000770, 0000000, 0031763, 0004000
    };

    static struct {
	short *data;
	short length;
	long  offset;
    } table[] = {
	{ boot,    sizeof(boot),    00000 },
	{ bitmap,  sizeof(bitmap),  01000 },
	{ direct1, sizeof(direct1), 01700 },
	{ direct2, sizeof(direct2), 06000 },
	{ NULL, 0, 0 }
    };

    // now write data from the table
    for (i = 0; table[i].length; i++) {
	lseek(fd, table[i].offset, SEEK_SET);
	if (write(fd, table[i].data, table[i].length) != table[i].length) return -1;
    }

    return 0;
}



//
// init XXDP file directory structures (based on XXDPv2.5)
//
static int xxdp_init (int fd)
{
    int i;

    static short dir1[] = { // offset 0001000
	0000002, 0000001, 0000007, 0000007
    };

    static short dir2[] = { // offset 0002000
	0000000, 0001002, 0000003, 0000011
    };

    static short dir3[] = { // offset 0003000
	0000004
    };

    static short dir4[] = { // offset 0004000
	0000005
    };

    static short dir5[] = { // offset 0005000
	0000006
    };

    static short dir6[] = { // offset 0007000
	0000000, 0000001, 0000074, 0000007, 0177777, 0177777, 0000377
    };

    static struct {
	short *data;
	short length;
	long  offset;
    } table[] = {
	{ dir1, sizeof(dir1), 01000 },
	{ dir2, sizeof(dir2), 02000 },
	{ dir3, sizeof(dir3), 03000 },
	{ dir4, sizeof(dir4), 04000 },
	{ dir5, sizeof(dir5), 05000 },
	{ dir6, sizeof(dir6), 07000 },
	{ NULL, 0, 0 }
    };

    // now write data from the table
    for (i = 0; table[i].length; i++) {
	lseek(fd, table[i].offset, SEEK_SET);
	if (write(fd, table[i].data, table[i].length) != table[i].length) return -1;
    }

    return 0;
}



//
// init a blank tape image (all zero)
//
static int zero_init (int fd)
{
    int i;
    char buf[BLOCKSIZE];

    // zero a block
    memset(buf, 0, sizeof(buf));

    // zero a whole tape
    lseek(fd, 0, SEEK_SET);
    for (i = 0; i < TAPESIZE; i++)
	if (write(fd, buf, sizeof(buf)) != sizeof(buf)) return -1;

    return 0;
}



//
// open a file for a unit
//
int fileopen (char *name,
	      int mode)
{
    int fd;

    // check if we can open any more units
    if (fpt >= NTU58) { error("no more units available"); return -1; }

    // save some data
    file[fpt].name = name;
    file[fpt].rflag = 1;
    if (mode == FILEWRITE) file[fpt].wflag = 1;
    if (mode == FILECREATE) file[fpt].wflag = file[fpt].cflag = 1;
    if (mode == FILERT11INIT) file[fpt].wflag = file[fpt].cflag = file[fpt].iflag = 1;
    if (mode == FILEXXDPINIT) file[fpt].wflag = file[fpt].cflag = file[fpt].xflag = 1;

    // open file if it exists
    if (file[fpt].wflag)
	fd = open(file[fpt].name, O_BINARY|O_RDWR, 0666);
    else
	fd = open(file[fpt].name, O_BINARY|O_RDONLY);

    // create file if it does not exist
    if (fd < 0 && file[fpt].cflag) fd = creat(file[fpt].name, 0666);
    if (fd < 0) { error("fileopen cannot open or create '%s'", file[fpt].name); return -2; }

    // store opened file information
    file[fpt].fd = fd;

    // zap tape if requested
    if (file[fpt].cflag) {
	if (!zero_init(fd)) {
	    info("initialize tape on '%s'", file[fpt].name);
	} else {
	    error("fileopen cannot init tape on '%s'", file[fpt].name);
	    return -3;
	}
    }

    // initialize RT-11 directory structure ?
    if (file[fpt].iflag) {
	if (!rt11_init(fd)) {
	    info("initialize RT-11 directory on '%s'", file[fpt].name);
	} else {
	    error("fileopen cannot init RT-11 filesystem on '%s'", file[fpt].name);
	    return -4;
	}
    }

    // initialize XXDP directory structure ?
    if (file[fpt].xflag) {
	if (!xxdp_init(fd)) {
	    info("initialize XXDP directory on '%s'", file[fpt].name);
	} else {
	    error("fileopen cannot init XXDP filesystem on '%s'", file[fpt].name);
	    return -5;
	}
    }

    // output some info...
    info("unit %d %c%c%c%c file '%s'",
	 fpt,
	 file[fpt].rflag ? 'r' : ' ',
	 file[fpt].wflag ? 'w' : ' ',
	 file[fpt].cflag ? 'c' : ' ',
	 file[fpt].iflag ? 'i' : file[fpt].xflag ? 'x' : ' ',
	 file[fpt].name);

    fpt++;
    return 0;
}



//
// check file unit OK
//
int fileunit (int unit)
{
    if (unit < 0 || unit >= NTU58 || file[unit].fd == -1) {
	error("bad unit %d", unit);
	return -1;
    }
    return 0;
}



//
// seek file (tape) position
//
int fileseek (int unit,
	      int size,
	      int block,
	      int offset)
{
    if (fileunit(unit)) return -1;

    if (block*size+offset > lseek(file[unit].fd, 0, SEEK_END)) return -2;

    if (lseek(file[unit].fd, block*size+offset, SEEK_SET) < 0) return -3;

    return 0;
}



//
// read bytes from the tape image file
//
int fileread (int unit,
	      unsigned char *buffer,
	      int count)
{
    if (fileunit(unit)) return -1;

    if (!file[unit].rflag) return -2;

    return read(file[unit].fd, buffer, count);
}



//
// write bytes to the tape image file
//
int filewrite (int unit,
	       unsigned char *buffer,
	       int count)
{
    if (fileunit(unit)) return -1;

    if (!file[unit].wflag) return -2;

    return write(file[unit].fd, buffer, count);
}



// the end
