//
// tu58 - Emulate a TU58 over a serial line
//
// Original (C) 1984 Dan Ts'o <Rockefeller Univ. Dept. of Neurobiology>
// Update (C) 2005 Don North <ak6dn_at_mindspring_dot_com>
//
// This is the TU58 emulation program written at Rockefeller Univ., Dept. of
// Neurobiology. We copyright (C) it and permit its use provided it is not
// sold to others. Originally written by Dan Ts'o circa 1984 or so.
//
// Extensively rewritten to work reliably under the CYGWIN unix subsystem.
// Has been tested up to 38.4Kbaud hosted on a 1GHz Pentium3 Win2K platform.
// Added native windows serial comm support compile option (for break detect).
// Added multithreaded supervisor for command parser and emulator restart.
// Added timing delay option to allow emulator to pass diagnostic ZTUUF0.BIN.
// Supports four (or more, compile option) logical TU58 devices per controller.
//
// v1.0b
// 12 Jul 2005 - donorth - Initial rewrite
// v1.1b
// 20 Feb 2006 - donorth - Fixed RT11 f/s init; added XXDP f/s init
// v1.2b
// 25 Feb 2006 - donorth - Updated -p arg to be number or string
//                       - Fixed some typos in help string
// v1.3b
// 28 Feb 2006 - donorth - Updated windoze serial line config
// v1.4b
// 25 May 2006 - donorth - Send INITs on initialization/restart
// v1.4c
// 25 Nov 2007 - donorth - Make <INIT><INIT> message be debug, not verbose
//                         Fix fileseek() to correctly detect LEOF --
//                         -- read/write no longer skip past LEOF
//


#include "common.h"
#include <getopt.h>


static char copyright[] = "(C) 2007 Don North <ak6dn" "@" "mindspring.com>, " \
                          "(C) 1984 Dan Ts'o <Rockefeller University>";

static char version[] = "tu58 tape emulator v1.4c";

static char port[32] = "1"; // default port number
static int speed = 9600; // default line speed

int verbose = 0; // set nonzero to output more info
int timing = 0; // set nonzero to add timing delays
int xoffen = 0; // set nonzero to honor XOFFs
int mrspen = 0; // set nonzero to enable MRSP mode
int debug = 0; // set nonzero for debug output



//
// print an info message and return
//
void info (char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    fprintf(stderr, "info: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
    return;
}



//
// print an error message and return
//
void error (char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    fprintf(stderr, "ERROR: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
    return;
}



//
// print an error message and die
//
void fatal (char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    fprintf(stderr, "FATAL: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
    exit(EXIT_FAILURE);
}



//
// main program
//
int main (int argc,
	  char **argv)
{
    int i;
    int n = 0;
    int errors = 0;

    // init file structures
    fileinit();

    // process command line options
    while ((i = getopt(argc, argv, "-dVvTtmxs:p:r:w:i:c:z:")) != EOF) {
	switch (i) {
	case 'p':  strcpy(port, optarg);  break;
	case 's':  speed = atoi(optarg);  break;
	case 'r':  fileopen(optarg, FILEREAD);  n++;  break;
	case 'w':  fileopen(optarg, FILEWRITE);  n++;  break;
	case 'c':  fileopen(optarg, FILECREATE);  n++;  break;
	case 'i':  fileopen(optarg, FILERT11INIT);  n++;  break;
	case 'z':  fileopen(optarg, FILEXXDPINIT);  n++;  break;
	case 'x':  xoffen = 1;  break;
	case 'm':  mrspen = 1;  break;
	case 'T':  timing = 2;  break;
	case 't':  timing = 1;  break;
	case 'd':  verbose = 1; debug = 1;  break;
	case 'v':  verbose = 1;  break;
	case 'V':  info("version is %s", version);  break;
	default:   errors++; break;
	}
    }

    // some debug info
    if (debug) info(copyright);

    // must have opened at least one unit
    if (n == 0) {
	error("no units were specified");
	errors++;
    }

    // any error seen, die and print out some help
    if (errors)
	fatal("illegal command line\n" \
	      "  %s\n" \
	      "  Usage: %s [-options] -[rwci] file1 ... -[rwci] file%d\n" \
	      "  Options: -V           output version string\n" \
	      "           -v           enable verbose output to terminal\n" \
	      "           -d           enable debug output to terminal\n" \
	      "           -x           honor XOFF flags\n" \
	      "           -m           enable MRSP mode\n" \
	      "           -t           add timing delays to spoof diagnostic into passing\n" \
	      "           -T           add timing delays to mimic a real TU58\n" \
	      "           -s BAUD      set line speed [1200..115200; default 9600]\n" \
	      "           -p PORT      set port to PORT [1..N or /dev/comN; default 1]\n" \
	      "           -r FILENAME  readonly drive\n" \
	      "           -w FILENAME  read/write drive\n" \
	      "           -c FILENAME  create new r/w drive, zero tape\n" \
	      "           -i FILENAME  create new r/w drive, initialize RT11 directory\n" \
	      "           -z FILENAME  create new r/w drive, initialize XXDP directory\n",
	      version, argv[0], NTU58-1);

    // give some info
    info("serial port %s at %d baud", port, speed);

    // comment about MRSP for now
    if (mrspen) {
	info("*** MRSP mode has NOT been fully tested - use with caution! ***");
    }

    // setup serial and console ports
    devinit(port, speed);
    coninit();
    
    // play TU58
    tu58();

    // restore serial and console ports
    conrestore();
    devrestore();

    // close files we opened
    fileclose();

    // and done
    return EXIT_SUCCESS;
}



// the end
